# authentic2 - versatile identity manager
# Copyright (C) 2010-2019 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# authentic2

from django.contrib.auth.middleware import AuthenticationMiddleware
from django.contrib.sessions.middleware import SessionMiddleware

from authentic2.utils import (good_next_url, same_origin, select_next_url,
                              user_can_change_password, login,
                              get_authentication_events, authenticate)


def test_good_next_url(db, rf, settings):
    request = rf.get('/', HTTP_HOST='example.net', **{'wsgi.url_scheme': 'https'})
    assert good_next_url(request, '/admin/')
    assert good_next_url(request, '/')
    assert good_next_url(request, 'https://example.net/')
    assert good_next_url(request, 'https://example.net:443/')
    assert not good_next_url(request, 'https://example.net:4443/')
    assert not good_next_url(request, 'http://example.net/')
    assert not good_next_url(request, 'https://google.com/')
    assert not good_next_url(request, '')
    assert not good_next_url(request, None)


def test_good_next_url_backends(rf, external_redirect):
    next_url, valid = external_redirect
    request = rf.get('/', HTTP_HOST='example.net', **{'wsgi.url_scheme': 'https'})
    if valid:
        assert good_next_url(request, next_url)
    else:
        assert not good_next_url(request, next_url)


def test_same_origin():
    assert same_origin('http://example.com/coin/', 'http://example.com/')
    assert same_origin('http://example.com/coin/', 'http://example.com:80/')
    assert same_origin('http://example.com:80/coin/', 'http://example.com/')
    assert same_origin('http://example.com:80/coin/', 'http://.example.com/')
    assert same_origin('http://example.com:80/coin/', '//example.com/')
    assert not same_origin('https://example.com:80/coin/', 'http://example.com/')
    assert not same_origin('http://example.com/coin/', 'http://bob.example.com/')
    assert same_origin('https://example.com/coin/', 'https://example.com:443/')
    assert not same_origin('https://example.com:34/coin/', 'https://example.com/')
    assert same_origin('https://example.com:34/coin/', '//example.com')
    assert not same_origin('https://example.com/coin/', '//example.com:34')
    assert same_origin('https://example.com:443/coin/', 'https://example.com/')
    assert same_origin('https://example.com:34/coin/', '//example.com')


def test_select_next_url(db, rf, settings):
    request = rf.get('/accounts/register/', data={'next': '/admin/'})
    assert select_next_url(request, '/') == '/admin/'
    request = rf.get('/accounts/register/', data={'next': 'http://example.com/'})
    assert select_next_url(request, '/') == '/'
    settings.A2_REDIRECT_WHITELIST = ['//example.com/']
    assert select_next_url(request, '/') == 'http://example.com/'


def test_user_can_change_password(simple_user, settings):
    assert user_can_change_password(user=simple_user) is True
    settings.A2_REGISTRATION_CAN_CHANGE_PASSWORD = False
    assert user_can_change_password(user=simple_user) is False


def test_get_authentication_events_hows(rf, simple_user):
    user = authenticate(username=simple_user.username, password=simple_user.username)
    request = rf.get('/login/')
    middleware = SessionMiddleware()
    middleware.process_request(request)
    middleware = AuthenticationMiddleware()
    middleware.process_request(request)
    assert 'password' not in [ev['how'] for ev in get_authentication_events(request)]
    login(request, user, 'password')
    assert 'password' in [ev['how'] for ev in get_authentication_events(request)]


def test_remember_cookie(rf):
    from authentic2.utils import get_remember_cookie, prepend_remember_cookie
    from django.http import HttpResponse

    request = rf.get('/')
    request.COOKIES['preferrence'] = '1 2'
    assert get_remember_cookie(request, 'preferrence') == [1, 2]
    request.COOKIES['preferrence'] = '1 2 3 4 5 6'
    assert get_remember_cookie(request, 'preferrence') == [1, 2, 3, 4, 5]
    request.COOKIES['preferrence'] = '1 2 3 4 x'
    assert get_remember_cookie(request, 'preferrence') == []
    request.COOKIES['preferrence'] = 'aaa'
    assert get_remember_cookie(request, 'preferrence') == []

    response = HttpResponse()
    request.COOKIES['preferrence'] = '1 2'
    prepend_remember_cookie(request, response, 'preferrence', 4)
    assert response.cookies['preferrence'].value == '4 1 2'

    response = HttpResponse()
    request.COOKIES['preferrence'] = '1 2 a'
    prepend_remember_cookie(request, response, 'preferrence', 4)
    assert response.cookies['preferrence'].value == '4'

    response = HttpResponse()
    request.COOKIES['preferrence'] = '1 2 3 4 5'
    prepend_remember_cookie(request, response, 'preferrence', 7)
    assert response.cookies['preferrence'].value == '7 1 2 3 4'
